/* hamachi.c: A Packet Engines GNIC-II Gigabit Ethernet driver for Linux. */
/*
	Written 1998-1999 by Donald Becker.

	This software may be used and distributed according to the terms
	of the GNU Public License, incorporated herein by reference.

	This driver is for the Packet Engines GNIC-II PCI Gigabit Ethernet
	adapter.

	The author may be reached as becker@tidalwave.net, or
	Donald Becker
	312 Severn Ave. #W302
	Annapolis MD 21403

	Support and updates available at
	http://www.tidalwave.net/~becker/hamachi.html
*/

static const char *version =
"hamachi.c:v0.11 8/21/99  Written by Donald Becker\n"
"  	http://www.tidalwave.net/~becker/hamachi.html\n";

/* A few user-configurable values. */

static int debug = 1;
#define hamachi_debug debug
static int max_interrupt_work = 40;
static int min_pci_latency = 32;
static int mtu = 0;

/* Set the copy breakpoint for the copy-only-tiny-frames scheme.
   Setting to > 1518 effectively disables this feature. */
static int rx_copybreak = 0;

/* A override for the hardware detection of bus width.
   Set to 1 to force 32 bit PCI bus detection.  Set to 4 to force 64 bit.
   Add 2 to disable parity detection.
*/
static int force32 = 0;

/* Used to pass the media type, etc.
   These exist for driver interoperability.
   No media types are currently defined.
*/
#define MAX_UNITS 8				/* More are supported, limit only on options */
static int options[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};
static int full_duplex[MAX_UNITS] = {-1, -1, -1, -1, -1, -1, -1, -1};

/* Operational parameters that are set at compile time. */

/* Keep the ring sizes a power of two for compile efficiency.
   The compiler will convert <unsigned>'%'<2^N> into a bit mask.
   Making the Tx ring too large decreases the effectiveness of channel
   bonding and packet priority.
   There are no ill effects from too-large receive rings. */
#define TX_RING_SIZE	64
#define TX_QUEUE_LEN	60		/* Limit ring entries actually used.  */
#define RX_RING_SIZE	128

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  (2*HZ)

/* Size of each temporary Rx buffer.  Add 8 if you do Rx checksumming! */
#define PKT_BUF_SZ		1544

#if !defined(__OPTIMIZE__)  ||  !defined(__KERNEL__)
#warning  You must compile this file with the correct options!
#warning  See the last lines of the source file.
#error You must compile this driver with "-O".
#endif

#include <linux/config.h>
#include <linux/version.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <asm/processor.h>		/* Processor type for cache alignment. */
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/unaligned.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

/* Kernel compatibility defines, some common to David Hind's PCMCIA package.
   This is only in the support-all-kernels source code. */

#define RUN_AT(x) (jiffies + (x))
/* Condensed bus+endian portability operations. */
#define virt_to_le32bus(addr)  cpu_to_le32(virt_to_bus(addr))
#define le32bus_to_virt(addr)  bus_to_virt(le32_to_cpu(addr))

#if (LINUX_VERSION_CODE >= 0x20100)
char kernel_version[] = UTS_RELEASE;
#else
#ifndef __alpha__
#define ioremap vremap
#define iounmap vfree
#endif
#endif
#if LINUX_VERSION_CODE < 0x20123
#define test_and_set_bit(val, addr) set_bit(val, addr)
#define le16_to_cpu(val) (val)
#define cpu_to_le16(val) (val)
#define le32_to_cpu(val) (val)
#define cpu_to_le32(val) (val)
#endif
#if LINUX_VERSION_CODE <= 0x20139
#define	net_device_stats enet_statistics
#else
#define NETSTATS_VER2
#endif
#if LINUX_VERSION_CODE < 0x20155
#include <linux/bios32.h>
#define PCI_SUPPORT_VER1
#else
#define PCI_SUPPORT_VER2
#endif
#if LINUX_VERSION_CODE < 0x20159
#define DEV_FREE_SKB(skb) dev_kfree_skb(skb, FREE_WRITE);
#else
#define DEV_FREE_SKB(skb) dev_kfree_skb(skb);
#endif
#if ! defined(CAP_NET_ADMIN)
#define capable(CAP_XXX) (suser())
#endif
#if ! defined(HAS_NETIF_QUEUE)
#define netif_wake_queue(dev)  mark_bh(NET_BH);
#endif

/*
				Theory of Operation

I. Board Compatibility

This device driver is designed for the Packet Engines "Hamachi"
Gigabit Ethernet chip.  The only PCA currently supported is the GNIC-II 64-bit
66Mhz PCI card.

II. Board-specific settings

No jumpers exist on the board.  The chip supports software correction of
various motherboard wiring errors, however this driver does not support
that feature.

III. Driver operation

IIIa. Ring buffers

The Hamachi uses a typical descriptor based bus-master architecture.
The descriptor list is similar to that used by the Digital Tulip.
This driver uses two statically allocated fixed-size descriptor lists
formed into rings by a branch from the final descriptor to the beginning of
the list.  The ring sizes are set at compile time by RX/TX_RING_SIZE.

This driver uses a zero-copy receive and transmit scheme similar my other
network drivers.
The driver allocates full frame size skbuffs for the Rx ring buffers at
open() time and passes the skb->data field to the Hamachi as receive data
buffers.  When an incoming frame is less than RX_COPYBREAK bytes long,
a fresh skbuff is allocated and the frame is copied to the new skbuff.
When the incoming frame is larger, the skbuff is passed directly up the
protocol stack and replaced by a newly allocated skbuff.

The RX_COPYBREAK value is chosen to trade-off the memory wasted by
using a full-sized skbuff for small frames vs. the copying costs of larger
frames.  Gigabit cards are typically used on generously configured machines
and the underfilled buffers have negligible impact compared to the benefit of
a single allocation size, so the default value of zero results in never
copying packets.

IIIb/c. Transmit/Receive Structure

The Rx and Tx descriptor structure are straight-forward, with no historical
baggage that must be explained.  Unlike the awkward DBDMA structure, there
are no unused fields or option bits that had only one allowable setting.

Two details should be noted about the descriptors: The chip supports both 32
bit and 64 bit address structures, and the length field is overwritten on
the receive descriptors.  The descriptor length is set in the control word
for each channel. The development driver uses 32 bit addresses only, however
64 bit addresses may be enabled for 64 bit architectures e.g. the Alpha.

IIId. Synchronization

This driver is very similar to my other network drivers.
The driver runs as two independent, single-threaded flows of control.  One
is the send-packet routine, which enforces single-threaded use by the
dev->tbusy flag.  The other thread is the interrupt handler, which is single
threaded by the hardware and other software.

The send packet thread has partial control over the Tx ring and 'dev->tbusy'
flag.  It sets the tbusy flag whenever it's queuing a Tx packet. If the next
queue slot is empty, it clears the tbusy flag when finished otherwise it sets
the 'hmp->tx_full' flag.

The interrupt handler has exclusive control over the Rx ring and records stats
from the Tx ring.  After reaping the stats, it marks the Tx queue entry as
empty by incrementing the dirty_tx mark. Iff the 'hmp->tx_full' flag is set, it
clears both the tx_full and tbusy flags.

IV. Notes

Thanks to Kim Stearns of Packet Engines for providing a pair of GNIC-II boards.

IVb. References

Hamachi Engineering Design Specification, 5/15/97
(Note: This version was marked "Confidential".)
http://cesdis.gsfc.nasa.gov/linux/misc/100mbps.html

IVc. Errata

None noted.
*/


/* The rest of these values should never change. */

static struct device *hamachi_probe1(int pci_bus, int pci_devfn,
									 struct device *dev, long ioaddr,
									 int irq, int table_idx, int fnd_cnt);
enum chip_capability_flags { CanHaveMII=1, };

enum pci_flags_bits {
	/* Set PCI command register bits before calling probe1(). */
	PCI_USES_IO=1, PCI_USES_MEM=2, PCI_USES_MASTER=4,
	/* Read and map a single PCI BAR. */
	PCI_ADDR0=0x10<<0, PCI_ADDR1=0x10<<1, PCI_ADDR2=0x10<<2, PCI_ADDR3=0x10<<3,
};
struct chip_info {
	const char *name;
	struct match_info {
		u16	vendor, device, device_mask, pad;
	} id;
	enum pci_flags_bits pci_flags;
	int io_size;				/* Needed for I/O region check or ioremap(). */
	int drv_flags;				/* Driver use, intended as capability flags. */
	struct device *(*probe1)(int pci_bus, int pci_devfn, struct device *dev,
							 long ioaddr, int irq, int table_idx, int fnd_cnt);
	/* Optional, called for suspend, resume and detach. */
	int (*pwr_event)(struct device *dev, int event);
} chip_tbl[] = {
	{"Packet Engines GNIC-II \"Hamachi\"", { 0x1318, 0x0911, 0xffff, 0,},
	 PCI_USES_MEM | PCI_USES_MASTER, 0x400, 0, hamachi_probe1,},
	{ 0,},
};

/* Offsets to the Hamachi registers.  Various sizes. */
enum hamachi_offsets {
	TxDMACtrl=0x00, TxCmd=0x04, TxStatus=0x06, TxPtr=0x08, TxCurPtr=0x10,
	RxDMACtrl=0x20, RxCmd=0x24, RxStatus=0x26, RxPtr=0x28, RxCurPtr=0x30,
	PCIClkMeas=0x060, MiscStatus=0x066, ChipRev=0x68, ChipReset=0x06B,
	LEDCtrl=0x06C, VirtualJumpers=0x06D,
	TxChecksum=0x074, RxChecksum=0x076,
	TxIntrCtrl=0x078, RxIntrCtrl=0x07C,
	InterruptEnable=0x080, InterruptClear=0x084, IntrStatus=0x088,
	EventStatus=0x08C,
	MACCnfg=0x0A0, FrameGap0=0x0A2, FrameGap1=0x0A4,
	/* See enum MII_offsets below. */
	MACCnfg2=0x0B0, RxDepth=0x0B8, FlowCtrl=0x0BC, MaxFrameSize=0x0CE,
	AddrMode=0x0D0, StationAddr=0x0D2,
	/* Gigabit AutoNegotiation. */
	ANCtrl=0x0E0, ANStatus=0x0E2, ANXchngCtrl=0x0E4, ANAdvertise=0x0E8,
	ANLinkPartnerAbility=0x0EA,
	EECmdStatus=0x0F0, EEData=0x0F1, EEAddr=0x0F2,
	FIFOcfg=0x0F8,
};

/* Offsets to the MII-mode registers. */
enum MII_offsets {
	MII_Cmd=0xA6, MII_Addr=0xA8, MII_Wr_Data=0xAA, MII_Rd_Data=0xAC,
	MII_Status=0xAE,
};

/* Bits in the interrupt status/mask registers. */
enum intr_status_bits {
	IntrRxDone=0x01, IntrRxPCIFault=0x02, IntrRxPCIErr=0x04,
	IntrTxDone=0x100, IntrTxPCIFault=0x200, IntrTxPCIErr=0x400,
	LinkChange=0x10000, NegotiationChange=0x20000, StatsMax=0x40000, };

/* The Hamachi Rx and Tx buffer descriptors. */
struct hamachi_desc {
	u16 length;			/* u16 == u_int16_t */
	u16 status;
#if ADDRLEN == 64
	u32 pad;
	u64 addr;
#else
	u32 addr;
#endif
};

/* Bits in hamachi_desc.status */
enum desc_status_bits {
	DescOwn=0x8000, DescEndPacket=0x4000, DescEndRing=0x2000, DescIntr=0x1000,
};

#define PRIV_ALIGN	15 	/* Required alignment mask */
struct hamachi_private {
	/* Descriptor rings first for alignment.  Tx requires a second descriptor
	   for status. */
	struct hamachi_desc rx_ring[RX_RING_SIZE];
	struct hamachi_desc tx_ring[TX_RING_SIZE];
	/* The addresses of receive-in-place skbuffs. */
	struct sk_buff* rx_skbuff[RX_RING_SIZE];
	/* The saved address of a sent-in-place packet/buffer, for skfree(). */
	struct sk_buff* tx_skbuff[TX_RING_SIZE];
	struct device *next_module;
	void *priv_addr;					/* Unaligned address for kfree */
	struct net_device_stats stats;
	struct timer_list timer;	/* Media selection timer. */
	/* Frequently used and paired value: keep adjacent for cache effect. */
	int chip_id;
	int in_interrupt;
	struct hamachi_desc *rx_head_desc;
	unsigned int cur_rx, dirty_rx;		/* Producer/consumer ring indices */
	unsigned int rx_buf_sz;				/* Based on MTU+slack. */
	unsigned int cur_tx, dirty_tx;
	unsigned int tx_full:1;				/* The Tx queue is full. */
	unsigned int full_duplex:1;			/* Full-duplex operation requested. */
	unsigned int duplex_lock:1;
	unsigned int medialock:1;			/* Do not sense media. */
	unsigned int default_port:4;		/* Last dev->if_port value. */
	/* MII transceiver section. */
	int mii_cnt;						/* MII device addresses. */
	u16 advertising;					/* NWay media advertisement */
	unsigned char phys[2];				/* MII device addresses. */
	u32 pad[4];							/* Used for 32-byte alignment */
};

#ifdef MODULE

#if LINUX_VERSION_CODE > 0x20115
MODULE_AUTHOR("Donald Becker <becker@cesdis.gsfc.nasa.gov>");
MODULE_DESCRIPTION("Packet Engines 'Hamachi' GNIC-II Gigabit Ethernet driver");
MODULE_PARM(max_interrupt_work, "i");
MODULE_PARM(min_pci_latency, "i");
MODULE_PARM(mtu, "i");
MODULE_PARM(debug, "i");
MODULE_PARM(rx_copybreak, "i");
MODULE_PARM(options, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(force32, "i");
#endif

#endif

static int read_eeprom(long ioaddr, int location);
static int mdio_read(long ioaddr, int phy_id, int location);
static void mdio_write(long ioaddr, int phy_id, int location, int value);
static int hamachi_open(struct device *dev);
#ifdef HAVE_PRIVATE_IOCTL
static int mii_ioctl(struct device *dev, struct ifreq *rq, int cmd);
#endif
#ifdef HAVE_CHANGE_MTU
static int change_mtu(struct device *dev, int new_mtu);
#endif
static void hamachi_timer(unsigned long data);
static void hamachi_tx_timeout(struct device *dev);
static void hamachi_init_ring(struct device *dev);
static int hamachi_start_xmit(struct sk_buff *skb, struct device *dev);
static void hamachi_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
static int hamachi_rx(struct device *dev);
static void hamachi_error(struct device *dev, int intr_status);
static int hamachi_close(struct device *dev);
static struct net_device_stats *hamachi_get_stats(struct device *dev);
static void set_rx_mode(struct device *dev);



/* A list of our installed devices, for removing the driver module. */
static struct device *root_hamachi_dev = NULL;

int hamachi_probe(struct device *dev)
{
	int cards_found = 0;
	int pci_index = 0;
	unsigned char pci_bus, pci_device_fn;

	if ( ! pcibios_present())
		return -ENODEV;

	for (;pci_index < 0xff; pci_index++) {
		u16 pci_command, new_command, vendor, device;
		int chip_idx;
		int irq;
		long pciaddr;
		long ioaddr;

		if (pcibios_find_class (PCI_CLASS_NETWORK_ETHERNET << 8,
								pci_index, &pci_bus, &pci_device_fn)
			!= PCIBIOS_SUCCESSFUL)
			break;
		pcibios_read_config_word(pci_bus, pci_device_fn,
								 PCI_VENDOR_ID, &vendor);
		pcibios_read_config_word(pci_bus, pci_device_fn,
								 PCI_DEVICE_ID, &device);

		for (chip_idx = 0; chip_tbl[chip_idx].id.vendor; chip_idx++)
			if (vendor == chip_tbl[chip_idx].id.vendor
				&& (device & chip_tbl[chip_idx].id.device_mask) ==
				chip_tbl[chip_idx].id.device)
				break;
		if (chip_tbl[chip_idx].id.vendor == 0) 		/* Compiled out! */
			continue;

		{
#if LINUX_VERSION_CODE > 0x20155  ||  PCI_SUPPORT_1
			struct pci_dev *pdev = pci_find_slot(pci_bus, pci_device_fn);
			pciaddr = pdev->base_address[0];
#ifdef __alpha__				/* Really "64 bit addrs" */
			pciaddr |= ((long)pdev->base_address[1]) << 32;
#endif
			irq = pdev->irq;
			printk(KERN_INFO "Found %s at PCI address %#lx, IRQ %d.\n",
				   chip_tbl[chip_idx].name, pciaddr, irq);
#else
			u32 pci_memaddr, pci_memaddr_hi = 0;
			u8 pci_irq_line;
			pcibios_read_config_byte(pci_bus, pci_device_fn,
									 PCI_INTERRUPT_LINE, &pci_irq_line);
			pcibios_read_config_dword(pci_bus, pci_device_fn,
									  PCI_BASE_ADDRESS_0, &pci_memaddr);
			pciaddr = pci_memaddr;
#ifdef __alpha__				/* Really "64 bit addrs" */
			pcibios_read_config_dword(pci_bus, pci_device_fn,
									  PCI_BASE_ADDRESS_1, &pci_memaddr_hi);
			pciaddr = pci_memaddr + ((long)pci_memaddr_hi<<32);
#endif
			irq = pci_irq_line;
			printk(KERN_INFO "Found %s at PCI address %#8.8x %8.8x, IRQ %d.\n",
				   chip_tbl[chip_idx].name, pci_memaddr_hi, pci_memaddr, irq);
#endif
		}

		if (hamachi_debug > 2)
			printk(KERN_INFO "Found %s at PCI address %#lx, IRQ %d.\n",
				   chip_tbl[chip_idx].name, pciaddr, irq);
		ioaddr = (long)ioremap(pciaddr & ~0xf, chip_tbl[chip_idx].io_size);

		if (ioaddr == 0) {
			printk(KERN_INFO "Failed to map PCI address %#lx.\n",
				   pciaddr);
			continue;
		}

		pcibios_read_config_word(pci_bus, pci_device_fn,
								 PCI_COMMAND, &pci_command);
		new_command = pci_command | PCI_COMMAND_MASTER|PCI_COMMAND_MEMORY;
		if (pci_command != new_command) {
			printk(KERN_INFO "  The PCI BIOS has not enabled the"
				   " device at %d/%d!  Updating PCI command %4.4x->%4.4x.\n",
				   pci_bus, pci_device_fn, pci_command, new_command);
			pcibios_write_config_word(pci_bus, pci_device_fn,
									  PCI_COMMAND, new_command);
		}

		dev = hamachi_probe1(pci_bus, pci_device_fn, dev, ioaddr, irq,
							 chip_idx, cards_found);

		if (dev) {
			u8 pci_latency;
			/* Disable timeouts (errata), check PCI latency values. */
			pcibios_write_config_byte(pci_bus, pci_device_fn, 0x40, 0);
			pcibios_write_config_byte(pci_bus, pci_device_fn, 0x41, 0);
			pcibios_read_config_byte(pci_bus, pci_device_fn,
									 PCI_LATENCY_TIMER, &pci_latency);
			if (pci_latency < min_pci_latency) {
				printk(KERN_INFO "  PCI latency timer (CFLT) is "
					   "unreasonably low at %d.  Setting to %d clocks.\n",
					   pci_latency, min_pci_latency);
				pcibios_write_config_byte(pci_bus, pci_device_fn,
										  PCI_LATENCY_TIMER, min_pci_latency);
			} else if (hamachi_debug > 1)
				printk(KERN_INFO "  PCI latency timer (CFLT) is %#x.\n",
					   pci_latency);
		}
		dev = 0;
		cards_found++;
	}

	return cards_found ? 0 : -ENODEV;
}

static struct device *hamachi_probe1(int pci_bus, int pci_devfn,
									 struct device *dev, long ioaddr,
									 int irq, int chip_idx, int fnd_cnt)
{
	static int did_version = 0;			/* Already printed version info. */
	struct hamachi_private *hmp;
	int option = 0, i;

	if (hamachi_debug > 0  &&  did_version++ == 0)
		printk(version);

	dev = init_etherdev(dev, 0);

	printk(KERN_INFO "%s: %s type %x at 0x%lx, ",
		   dev->name, chip_tbl[chip_idx].name, readl(ioaddr + ChipRev),
		   ioaddr);

	for (i = 0; i < 6; i++)
		dev->dev_addr[i] = 1 ? read_eeprom(ioaddr, 4 + i)
			: readb(ioaddr + StationAddr + i);
	for (i = 0; i < 5; i++)
			printk("%2.2x:", dev->dev_addr[i]);
	printk("%2.2x, IRQ %d.\n", dev->dev_addr[i], irq);

	i = readb(ioaddr + PCIClkMeas);
	printk(KERN_INFO "%s:  %d-bit %d Mhz PCI bus (%d), Virtual Jumpers "
		   "%2.2x, LPA %4.4x.\n",
		   dev->name, readw(ioaddr + MiscStatus) & 1 ? 64 : 32,
		   i ? 2000/(i&0x7f) : 0, i&0x7f, (int)readb(ioaddr + VirtualJumpers),
		   readw(ioaddr + ANLinkPartnerAbility));

	/* If the bus size is misidentified, do the following. */
	if (force32)
		writeb(force32, ioaddr + VirtualJumpers);

	/* Hmmm, do we really need to reset the chip???. */
	writeb(1, ioaddr + ChipReset);

	dev->base_addr = ioaddr;
	dev->irq = irq;

	/* Make certain the descriptor lists are aligned. */
	{
		void *mem = kmalloc(sizeof(*hmp) + PRIV_ALIGN, GFP_KERNEL);
		hmp =  (void *)(((long)mem + PRIV_ALIGN) & ~PRIV_ALIGN);
		dev->priv = hmp;
		hmp->priv_addr = mem;
	}
	memset(hmp, 0, sizeof(*hmp));

	hmp->next_module = root_hamachi_dev;
	root_hamachi_dev = dev;

	hmp->chip_id = chip_idx;

	if (fnd_cnt < MAX_UNITS) {
		if (full_duplex[fnd_cnt] > 0)
			hmp->full_duplex = 1;
		option = options[fnd_cnt];
	}
	if (dev->mem_start)
		option = dev->mem_start;

	/* The lower four bits are the media type. */
	if (option > 0) {
		if (option & 0x200)
			hmp->full_duplex = 1;
		hmp->default_port = option & 15;
		if (hmp->default_port)
			hmp->medialock = 1;
	}
	if (hmp->full_duplex)
		hmp->duplex_lock = 1;

	/* The Hamachi-specific entries in the device structure. */
	dev->open = &hamachi_open;
	dev->hard_start_xmit = &hamachi_start_xmit;
	dev->stop = &hamachi_close;
	dev->get_stats = &hamachi_get_stats;
	dev->set_multicast_list = &set_rx_mode;
#ifdef HAVE_PRIVATE_IOCTL
	dev->do_ioctl = &mii_ioctl;
#endif
#ifdef HAVE_CHANGE_MTU
  dev->change_mtu = change_mtu;
#endif
	if (mtu)
		dev->mtu = mtu;

	if (chip_tbl[hmp->chip_id].drv_flags & CanHaveMII) {
		int phy, phy_idx = 0;
		for (phy = 0; phy < 32 && phy_idx < 4; phy++) {
			int mii_status = mdio_read(ioaddr, phy, 1);
			if (mii_status != 0xffff  &&
				mii_status != 0x0000) {
				hmp->phys[phy_idx++] = phy;
				hmp->advertising = mdio_read(ioaddr, phy, 4);
				printk(KERN_INFO "%s: MII PHY found at address %d, status "
					   "0x%4.4x advertising %4.4x.\n",
					   dev->name, phy, mii_status, hmp->advertising);
			}
		}
		hmp->mii_cnt = phy_idx;
	}
#ifdef notyet	
	/* Disable PCI Parity Error (0x02) or PCI 64 Bit (0x01) for miswired
	   motherboards. */
	if (readb(ioaddr + VirtualJumpers) != 0x30)
		writeb(0x33, ioaddr + VirtualJumpers)
#endif
	/* Configure gigabit autonegotiation. */
	writew(0x0400, ioaddr + ANXchngCtrl);	/* Enable legacy links. */
	writew(0x08e0, ioaddr + ANAdvertise);	/* Set our advertise word. */
	writew(0x1000, ioaddr + ANCtrl);		/* Enable negotiation */

	return dev;
}

static int read_eeprom(long ioaddr, int location)
{
	int bogus_cnt = 1000;

	writew(location, ioaddr + EEAddr);
	writeb(0x02, ioaddr + EECmdStatus);
	while ((readb(ioaddr + EECmdStatus) & 0x40)  && --bogus_cnt > 0)
		;
	if (hamachi_debug > 5)
		printk("   EEPROM status is %2.2x after %d ticks.\n",
			   (int)readb(ioaddr + EECmdStatus), 1000- bogus_cnt);
	return readb(ioaddr + EEData);
}

/* MII Managemen Data I/O accesses.
   These routines assume the MDIO controller is idle, and do not exit until
   the command is finished. */

static int mdio_read(long ioaddr, int phy_id, int location)
{
	int i;

	writew((phy_id<<8) + location, ioaddr + MII_Addr);
	writew(1, ioaddr + MII_Cmd);
	for (i = 10000; i >= 0; i--)
		if ((readw(ioaddr + MII_Status) & 1) == 0)
			break;
	return readw(ioaddr + MII_Rd_Data);
}

static void mdio_write(long ioaddr, int phy_id, int location, int value)
{
	int i;

	writew((phy_id<<8) + location, ioaddr + MII_Addr);
	writew(value, ioaddr + MII_Wr_Data);

	/* Wait for the command to finish. */
	for (i = 10000; i >= 0; i--)
		if ((readw(ioaddr + MII_Status) & 1) == 0)
			break;
	return;
}


static int hamachi_open(struct device *dev)
{
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int i;

	/* Do we need to reset the chip??? */

	if (request_irq(dev->irq, &hamachi_interrupt, SA_SHIRQ, dev->name, dev))
		return -EAGAIN;

	if (hamachi_debug > 1)
		printk(KERN_DEBUG "%s: hamachi_open() irq %d.\n",
			   dev->name, dev->irq);

	MOD_INC_USE_COUNT;

	hamachi_init_ring(dev);

#if ADDRLEN == 64
	writel(virt_to_bus(hmp->rx_ring), ioaddr + RxPtr);
	writel(virt_to_bus(hmp->rx_ring) >> 32, ioaddr + RxPtr + 4);
	writel(virt_to_bus(hmp->tx_ring), ioaddr + TxPtr);
	writel(virt_to_bus(hmp->tx_ring) >> 32, ioaddr + TxPtr + 4);
#else
	writel(virt_to_bus(hmp->rx_ring), ioaddr + RxPtr);
	writel(virt_to_bus(hmp->tx_ring), ioaddr + TxPtr);
#endif

	for (i = 0; i < 6; i++)
		writeb(dev->dev_addr[i], ioaddr + StationAddr + i);

	/* Initialize other registers: with so many this eventually this will
	   converted to an offset/value list. */
	/* Configure the FIFO for 512K external, 16K used for Tx. */
	writew(0x0028, ioaddr + FIFOcfg);

	if (dev->if_port == 0)
		dev->if_port = hmp->default_port;

	dev->tbusy = 0;
	dev->interrupt = 0;
	hmp->in_interrupt = 0;

	/* Setting the Rx mode will start the Rx process. */
	/* We are always in full-duplex mode with gigabit! */
	hmp->full_duplex = 1;
	writew(0x0001, ioaddr + RxChecksum); /* Enable Rx IP partial checksum. */
	writew(0x8000, ioaddr + MACCnfg); /* Soft reset the MAC */
	writew(0x215F, ioaddr + MACCnfg);
	writew(0x000C, ioaddr + FrameGap0); /* 0060/4060 for non-MII 10baseT */
	writew(0x1018, ioaddr + FrameGap1);
	writew(0x2780, ioaddr + MACCnfg2); /* Upper 16 bits control LEDs. */
	/* Enable automatic generation of flow control frames, period 0xffff. */
	writel(0x0030FFFF, ioaddr + FlowCtrl);
	writew(dev->mtu+19, ioaddr + MaxFrameSize); 	/* hmp->rx_buf_sz ??? */

	/* Enable legacy links. */
	writew(0x0400, ioaddr + ANXchngCtrl);	/* Enable legacy links. */
	/* Initial Link LED to blinking red. */
	writeb(0x03, ioaddr + LEDCtrl);

	/* Configure interrupt mitigation.  This has a great effect on
	   performance, so systems tuning should start here!. */
	writel(0x00080000, ioaddr + TxIntrCtrl);
	writel(0x00000020, ioaddr + RxIntrCtrl);

	set_rx_mode(dev);

	dev->start = 1;

	/* Enable interrupts by setting the interrupt mask. */
	writel(0x80878787, ioaddr + InterruptEnable);
	writew(0x0000, ioaddr + EventStatus);	/* Clear non-interrupting events */

	/* Configure and start the DMA channels. */
	/* Burst sizes are in the low three bits: size = 4<<(val&7) */
#if ADDRLEN == 64
	writew(0x0055, ioaddr + RxDMACtrl); 		/* 128 dword bursts */
	writew(0x0055, ioaddr + TxDMACtrl);
#else
	writew(0x0015, ioaddr + RxDMACtrl);
	writew(0x0015, ioaddr + TxDMACtrl);
#endif
	writew(1, dev->base_addr + RxCmd);

	if (hamachi_debug > 2) {
		printk(KERN_DEBUG "%s: Done hamachi_open(), status: Rx %x Tx %x.\n",
			   dev->name, readw(ioaddr + RxStatus), readw(ioaddr + TxStatus));
	}
	/* Set the timer to check for link beat. */
	init_timer(&hmp->timer);
	hmp->timer.expires = RUN_AT((24*HZ)/10);			/* 2.4 sec. */
	hmp->timer.data = (unsigned long)dev;
	hmp->timer.function = &hamachi_timer;				/* timer handler */
	add_timer(&hmp->timer);

	return 0;
}

static void hamachi_timer(unsigned long data)
{
	struct device *dev = (struct device *)data;
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int next_tick = 10*HZ;

	if (hamachi_debug > 2) {
		printk(KERN_INFO "%s: Hamchi Autonegotiation status %4.4x, LPA "
			   "%4.4x.\n", dev->name, readw(ioaddr + ANStatus),
			   readw(ioaddr + ANLinkPartnerAbility));
		printk(KERN_INFO "%s: Autonegotiation regs %4.4x %4.4x %4.4x "
			   "%4.4x %4.4x %4.4x.\n", dev->name,
		       readw(ioaddr + 0x0e0),
			   readw(ioaddr + 0x0e2),
			   readw(ioaddr + 0x0e4),
			   readw(ioaddr + 0x0e6),
			   readw(ioaddr + 0x0e8),
			   readw(ioaddr + 0x0eA));
	}
	/* This has a small false-trigger window. */
	if (test_bit(0, (void*)&dev->tbusy) &&
		(jiffies - dev->trans_start) > TX_TIMEOUT
		&& hmp->cur_tx - hmp->dirty_tx > 1) {
		hamachi_tx_timeout(dev);
	}
	/* We could do something here... nah. */
	hmp->timer.expires = RUN_AT(next_tick);
	add_timer(&hmp->timer);
}

static void hamachi_tx_timeout(struct device *dev)
{
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	long ioaddr = dev->base_addr;

	printk(KERN_WARNING "%s: Hamachi transmit timed out, status %8.8x,"
		   " resetting...\n", dev->name, (int)readw(ioaddr + TxStatus));

#ifndef __alpha__
	{
		int i;
		printk(KERN_DEBUG "  Rx ring %8.8x: ", (int)hmp->rx_ring);
		for (i = 0; i < RX_RING_SIZE; i++)
			printk(" %8.8x", (unsigned int)hmp->rx_ring[i].status);
		printk("\n"KERN_DEBUG"  Tx ring %8.8x: ", (int)hmp->tx_ring);
		for (i = 0; i < TX_RING_SIZE; i++)
			printk(" %4.4x", hmp->tx_ring[i].status);
		printk("\n");
	}
#endif

	/* Perhaps we should reinitialize the hardware here. */
	dev->if_port = 0;
	/* Stop and restart the chip's Tx processes . */

	/* Trigger an immediate transmit demand. */
	writew(2, dev->base_addr + TxCmd);
	writew(1, dev->base_addr + TxCmd);
	writew(1, dev->base_addr + RxCmd);

	dev->trans_start = jiffies;
	hmp->stats.tx_errors++;
	return;
}


/* Initialize the Rx and Tx rings, along with various 'dev' bits. */
static void hamachi_init_ring(struct device *dev)
{
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	int i;

	hmp->tx_full = 0;
	hmp->cur_rx = hmp->cur_tx = 0;
	hmp->dirty_rx = hmp->dirty_tx = 0;

	hmp->rx_buf_sz = (dev->mtu <= 1500 ? PKT_BUF_SZ : dev->mtu + 40);
	hmp->rx_head_desc = &hmp->rx_ring[0];

	for (i = 0; i < RX_RING_SIZE; i++) {
		struct sk_buff *skb;

		hmp->rx_ring[i].length = hmp->rx_buf_sz;

		skb = dev_alloc_skb(hmp->rx_buf_sz);
		hmp->rx_skbuff[i] = skb;
		if (skb) {
			skb->dev = dev;			/* Mark as being used by this device. */
			skb_reserve(skb, 2);	/* 16 byte align the IP header. */
			hmp->rx_ring[i].addr = virt_to_bus(skb->tail);
			hmp->rx_ring[i].status = DescOwn | DescEndPacket | DescIntr;
		} else {
			hmp->rx_ring[i].status = 0;
			if (hmp->dirty_rx == 0) 		/* No previous alloc failure */
				hmp->dirty_rx = (unsigned int)(0 - RX_RING_SIZE);
		}
	}
	/* Mark the last entry as wrapping the ring. */
	hmp->rx_ring[i-1].status |= DescEndRing;

	for (i = 0; i < TX_RING_SIZE; i++) {
		hmp->tx_skbuff[i] = 0;
		hmp->tx_ring[i].status = 0;
	}
	return;
}

static int hamachi_start_xmit(struct sk_buff *skb, struct device *dev)
{
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	unsigned entry;

	/* Block a timer-based transmit from overlapping.  This could better be
	   done with atomic_swap(1, dev->tbusy), but set_bit() works as well. */
	if (test_and_set_bit(0, (void*)&dev->tbusy) != 0) {
		if (jiffies - dev->trans_start > TX_TIMEOUT)
			hamachi_tx_timeout(dev);
		return 1;
	}

	/* Caution: the write order is important here, set the field
	   with the "ownership" bits last. */

	/* Calculate the next Tx descriptor entry. */
	entry = hmp->cur_tx % TX_RING_SIZE;

	hmp->tx_skbuff[entry] = skb;

	hmp->tx_ring[entry].addr = virt_to_bus(skb->data);
	hmp->tx_ring[entry].length = skb->len;
	if (entry >= TX_RING_SIZE-1)		 /* Wrap ring */
		hmp->tx_ring[entry].status = DescOwn|DescEndPacket|DescEndRing|DescIntr;
	else
		hmp->tx_ring[entry].status = DescOwn|DescEndPacket;
	hmp->cur_tx++;

	/* Non-x86 Todo: explicitly flush cache lines here. */

	/* Wake the potentially-idle transmit channel. */
	writew(1, dev->base_addr + TxCmd);

	if (hmp->cur_tx - hmp->dirty_tx < TX_QUEUE_LEN - 1)
		clear_bit(0, (void*)&dev->tbusy);		/* Typical path */
	else
		hmp->tx_full = 1;
	dev->trans_start = jiffies;

	if (hamachi_debug > 4) {
		printk(KERN_DEBUG "%s: Hamachi transmit frame #%d length %d queued "
			   "in slot %d.\n", dev->name, hmp->cur_tx, (int)skb->len, entry);
	}
	return 0;
}

/* The interrupt handler does all of the Rx thread work and cleans up
   after the Tx thread. */
static void hamachi_interrupt(int irq, void *dev_instance, struct pt_regs *rgs)
{
	struct device *dev = (struct device *)dev_instance;
	struct hamachi_private *hmp;
	long ioaddr, boguscnt = max_interrupt_work;

#ifndef final_version			/* Can never occur. */
	if (dev == NULL) {
		printk (KERN_ERR "hamachi_interrupt(): irq %d for unknown device.\n", irq);
		return;
	}
#endif

	ioaddr = dev->base_addr;
	hmp = (struct hamachi_private *)dev->priv;
	if (test_and_set_bit(0, (void*)&hmp->in_interrupt)) {
		printk(KERN_ERR "%s: Re-entering the interrupt handler.\n", dev->name);
		hmp->in_interrupt = 0;	/* Avoid future hang on bug */
		return;
	}

	do {
		u32 intr_status = readl(ioaddr + InterruptClear);

		if (hamachi_debug > 4)
			printk(KERN_DEBUG "%s: Hamachi interrupt, status %4.4x.\n",
				   dev->name, intr_status);

		if (intr_status == 0)
			break;

		if (intr_status & IntrRxDone)
			hamachi_rx(dev);

		for (; hmp->cur_tx - hmp->dirty_tx > 0; hmp->dirty_tx++) {
			int entry = hmp->dirty_tx % TX_RING_SIZE;
			if (hmp->tx_ring[entry].status == 0)
				break;
			/* Free the original skb. */
			DEV_FREE_SKB(hmp->tx_skbuff[entry]);
			hmp->tx_skbuff[entry] = 0;
			hmp->stats.tx_packets++;
		}
		if (hmp->tx_full && dev->tbusy
			&& hmp->cur_tx - hmp->dirty_tx < TX_QUEUE_LEN - 4) {
			/* The ring is no longer full, clear tbusy. */
			hmp->tx_full = 0;
			clear_bit(0, (void*)&dev->tbusy);
			netif_wake_queue(dev);
		}

		/* Abnormal error summary/uncommon events handlers. */
		if (intr_status &
			(IntrTxPCIFault | IntrTxPCIErr | IntrRxPCIFault | IntrRxPCIErr |
			 LinkChange | NegotiationChange | StatsMax))
			hamachi_error(dev, intr_status);

		if (--boguscnt < 0) {
			printk(KERN_WARNING "%s: Too much work at interrupt, status=0x%4.4x.\n",
				   dev->name, intr_status);
			break;
		}
	} while (1);

	if (hamachi_debug > 3)
		printk(KERN_DEBUG "%s: exiting interrupt, status=%#4.4x.\n",
			   dev->name, readl(ioaddr + IntrStatus));

#ifndef final_version
	/* Code that should never be run!  Perhaps remove after testing.. */
	{
		static int stopit = 10;
		if (dev->start == 0  &&  --stopit < 0) {
			printk(KERN_ERR "%s: Emergency stop, looping startup interrupt.\n",
				   dev->name);
			free_irq(irq, dev);
		}
	}
#endif

	dev->interrupt = 0;
	clear_bit(0, (void*)&hmp->in_interrupt);
	return;
}

/* This routine is logically part of the interrupt handler, but separated
   for clarity and better register allocation. */
static int hamachi_rx(struct device *dev)
{
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	int entry = hmp->cur_rx % RX_RING_SIZE;
	int boguscnt = 20;

	if (hamachi_debug > 4) {
		printk(KERN_DEBUG " In hamachi_rx(), entry %d status %4.4x.\n",
			   entry, hmp->rx_ring[entry].status);
	}

	/* If EOP is set on the next entry, it's a new packet. Send it up. */
	while ( ! (hmp->rx_head_desc->status & DescOwn)) {
		struct hamachi_desc *desc = hmp->rx_head_desc;
		u16 desc_status = desc->status;
		int data_size = desc->length;
		u8 *buf_addr = bus_to_virt(desc->addr);
		s32 frame_status = get_unaligned((s32*)&(buf_addr[data_size - 12]));

		if (hamachi_debug > 4)
			printk(KERN_DEBUG "  hamachi_rx() status was %8.8x.\n",
				   frame_status);
		if (--boguscnt < 0)
			break;
		if ( ! (desc_status & DescEndPacket)) {
			printk(KERN_WARNING "%s: Oversized Ethernet frame spanned "
				   "multiple buffers, entry %#x length %d status %4.4x!\n",
				   dev->name, hmp->cur_rx, data_size, desc_status);
			printk(KERN_WARNING "%s: Oversized Ethernet frame %p vs %p.\n",
				   dev->name, desc, &hmp->rx_ring[hmp->cur_rx % RX_RING_SIZE]);
			printk(KERN_WARNING "%s: Oversized Ethernet frame -- next status %x/%x last status %x.\n",
				   dev->name,
				   hmp->rx_ring[(hmp->cur_rx+1) % RX_RING_SIZE].status,
				   hmp->rx_ring[(hmp->cur_rx+1) % RX_RING_SIZE].length,
				   hmp->rx_ring[(hmp->cur_rx-1) % RX_RING_SIZE].status);
			hmp->stats.rx_length_errors++;
		} /* else  Omit for prototype errata??? */
		if (frame_status & 0x00380000) {
			/* There was a error. */
			if (hamachi_debug > 2)
				printk(KERN_DEBUG "  hamachi_rx() Rx error was %8.8x.\n",
					   frame_status);
			hmp->stats.rx_errors++;
			if (frame_status & 0x00600000) hmp->stats.rx_length_errors++;
			if (frame_status & 0x00080000) hmp->stats.rx_frame_errors++;
			if (frame_status & 0x00100000) hmp->stats.rx_crc_errors++;
			if (frame_status < 0) hmp->stats.rx_dropped++;
		} else {
			struct sk_buff *skb;
			u16 pkt_len = frame_status - 4;	/* Implicit truncate, omit CRC */

#ifndef final_version
			if (hamachi_debug > 4)
				printk(KERN_DEBUG "  hamachi_rx() normal Rx pkt length %d"
					   " of %d, bogus_cnt %d.\n",
					   pkt_len, data_size, boguscnt);
			if (hamachi_debug > 5)
				printk(KERN_DEBUG"%s:  rx status %8.8x %8.8x %8.8x %8.8x %8.8x.\n",
					   dev->name,
					   *(s32*)&(buf_addr[data_size - 20]),
					   *(s32*)&(buf_addr[data_size - 16]),
					   *(s32*)&(buf_addr[data_size - 12]),
					   *(s32*)&(buf_addr[data_size - 8]),
					   *(s32*)&(buf_addr[data_size - 4]));
#endif
			/* Check if the packet is long enough to accept without copying
			   to a minimally-sized skbuff. */
			if (pkt_len < rx_copybreak
				&& (skb = dev_alloc_skb(pkt_len + 2)) != NULL) {
				skb->dev = dev;
				skb_reserve(skb, 2);	/* 16 byte align the IP header */
				eth_copy_and_sum(skb, bus_to_virt(desc->addr), pkt_len, 0);
				skb_put(skb, pkt_len);
			} else {
				char *temp = skb_put(skb = hmp->rx_skbuff[entry], pkt_len);
				hmp->rx_skbuff[entry] = NULL;
#ifndef final_version				/* Remove after testing. */
				if (bus_to_virt(desc->addr) != temp)
					printk(KERN_ERR "%s: Internal fault: The skbuff addresses "
						   "do not match in hamachi_rx: %p vs. %p / %p.\n",
						   dev->name, bus_to_virt(desc->addr),
						   skb->head, temp);
#endif
			}
			skb->protocol = eth_type_trans(skb, dev);
			/* Note: checksum -> skb->ip_summed = CHECKSUM_UNNECESSARY; */
			netif_rx(skb);
			dev->last_rx = jiffies;
			hmp->stats.rx_packets++;
		}
		entry = (++hmp->cur_rx) % RX_RING_SIZE;
		hmp->rx_head_desc = &hmp->rx_ring[entry];
	}

	/* Refill the Rx ring buffers. */
	for (; hmp->cur_rx - hmp->dirty_rx > 0; hmp->dirty_rx++) {
		struct sk_buff *skb;
		entry = hmp->dirty_rx % RX_RING_SIZE;
		if (hmp->rx_skbuff[entry] == NULL) {
			skb = dev_alloc_skb(hmp->rx_buf_sz);
			hmp->rx_skbuff[entry] = skb;
			if (skb == NULL)
				break;			/* Better luck next round. */
			skb->dev = dev;			/* Mark as being used by this device. */
			skb_reserve(skb, 2);	/* Align IP on 16 byte boundaries */
			hmp->rx_ring[entry].addr = virt_to_bus(skb->tail);
		}
		hmp->rx_ring[entry].length = hmp->rx_buf_sz;
		if (entry >= RX_RING_SIZE - 1)
			hmp->rx_ring[entry].status =
				DescOwn | DescEndPacket | DescEndRing | DescIntr;
		else
			hmp->rx_ring[entry].status = DescOwn | DescEndPacket | DescIntr;
	}

	/* Restart Rx engine if stopped. */
	writew(1, dev->base_addr + RxCmd);
	return 0;
}

/* This is more properly named "uncommon interrupt events", as it covers more
   than just errors. */
static void hamachi_error(struct device *dev, int intr_status)
{
	long ioaddr = dev->base_addr;
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;

	if (intr_status & (LinkChange|NegotiationChange)) {
		if (hamachi_debug > 1)
			printk(KERN_INFO "%s: Link changed: AutoNegotiation Ctrl"
				   " %4.4x, Status %4.4x %4.4x Intr status %4.4x.\n",
				   dev->name, readw(ioaddr + 0x0E0), readw(ioaddr + 0x0E2),
				   readw(ioaddr + ANLinkPartnerAbility),
				   readl(ioaddr + IntrStatus));
		if (readw(ioaddr + ANStatus) & 0x20)
			writeb(0x01, ioaddr + LEDCtrl);
		else
			writeb(0x03, ioaddr + LEDCtrl);
	}
	if (intr_status & StatsMax) {
		hamachi_get_stats(dev);
		/* Read the overflow bits to clear. */
		readl(ioaddr + 0x36C);
		readl(ioaddr + 0x3F0);
	}
	if ((intr_status & ~(LinkChange|StatsMax|NegotiationChange))
		&& hamachi_debug)
		printk(KERN_ERR "%s: Something Wicked happened! %4.4x.\n",
			   dev->name, intr_status);
	/* Hmmmmm, it's not clear how to recover from PCI faults. */
	if (intr_status & (IntrTxPCIErr | IntrTxPCIFault))
		hmp->stats.tx_fifo_errors++;
	if (intr_status & (IntrRxPCIErr | IntrRxPCIFault))
		hmp->stats.rx_fifo_errors++;
}

static int hamachi_close(struct device *dev)
{
	long ioaddr = dev->base_addr;
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;
	int i;

	dev->start = 0;
	dev->tbusy = 1;

	if (hamachi_debug > 1) {
		printk(KERN_DEBUG "%s: Shutting down ethercard, status was Tx %4.4x Rx %4.4x Int %2.2x.\n",
			   dev->name, readw(ioaddr + TxStatus),
			   readw(ioaddr + RxStatus), readl(ioaddr + IntrStatus));
		printk(KERN_DEBUG "%s: Queue pointers were Tx %d / %d,  Rx %d / %d.\n",
			   dev->name, hmp->cur_tx, hmp->dirty_tx, hmp->cur_rx, hmp->dirty_rx);
	}

	/* Disable interrupts by clearing the interrupt mask. */
	writel(0x0000, ioaddr + InterruptEnable);

	/* Stop the chip's Tx and Rx processes. */
	writel(2, ioaddr + RxCmd);
	writew(2, ioaddr + TxCmd);

	del_timer(&hmp->timer);

#ifdef __i386__
	if (hamachi_debug > 2) {
		printk("\n"KERN_DEBUG"  Tx ring at %8.8x:\n",
			   (int)virt_to_bus(hmp->tx_ring));
		for (i = 0; i < TX_RING_SIZE; i++)
			printk(" %c #%d desc. %4.4x %4.4x %8.8x.\n",
				   readl(ioaddr + TxCurPtr) == (long)&hmp->tx_ring[i] ? '>' : ' ',
				   i, hmp->tx_ring[i].length,
				   hmp->tx_ring[i].status, hmp->tx_ring[i].addr);
		printk("\n"KERN_DEBUG "  Rx ring %8.8x:\n",
			   (int)virt_to_bus(hmp->rx_ring));
		for (i = 0; i < RX_RING_SIZE; i++) {
			printk(KERN_DEBUG " %c #%d desc. %4.4x %4.4x %8.8x\n",
				   readl(ioaddr + RxCurPtr) == (long)&hmp->rx_ring[i] ? '>' : ' ',
				   i, hmp->rx_ring[i].length,
				   hmp->rx_ring[i].status, hmp->rx_ring[i].addr);
			if (hamachi_debug > 6) {
				if (*(u8*)hmp->rx_ring[i].addr != 0x69) {
					int j;
					for (j = 0; j < 0x50; j++)
						printk(" %4.4x", ((u16*)hmp->rx_ring[i].addr)[j]);
					printk("\n");
				}
			}
		}
	}
#endif /* __i386__ debugging only */

	free_irq(dev->irq, dev);

	/* Free all the skbuffs in the Rx queue. */
	for (i = 0; i < RX_RING_SIZE; i++) {
		hmp->rx_ring[i].status = 0;
		hmp->rx_ring[i].addr = 0xBADF00D0; /* An invalid address. */
		if (hmp->rx_skbuff[i]) {
#if LINUX_VERSION_CODE < 0x20100
			hmp->rx_skbuff[i]->free = 1;
#endif
			DEV_FREE_SKB(hmp->rx_skbuff[i]);
		}
		hmp->rx_skbuff[i] = 0;
	}
	for (i = 0; i < TX_RING_SIZE; i++) {
		if (hmp->tx_skbuff[i])
			DEV_FREE_SKB(hmp->tx_skbuff[i]);
		hmp->tx_skbuff[i] = 0;
	}

	writeb(0x00, ioaddr + LEDCtrl);

	MOD_DEC_USE_COUNT;

	return 0;
}

static struct enet_statistics *hamachi_get_stats(struct device *dev)
{
	long ioaddr = dev->base_addr;
	struct hamachi_private *hmp = (struct hamachi_private *)dev->priv;

	/* We should lock this segment of code for SMP eventually, although
	   the vulnerability window is very small and statistics are
	   non-critical. */
#if LINUX_VERSION_CODE >= 0x20119
	hmp->stats.rx_bytes += readl(ioaddr + 0x330); /* Total Uni+Brd+Multi */
	hmp->stats.tx_bytes += readl(ioaddr + 0x3B0); /* Total Uni+Brd+Multi */
#endif
	hmp->stats.multicast		+= readl(ioaddr + 0x320); /* Multicast Rx */

	hmp->stats.rx_length_errors	+= readl(ioaddr + 0x368); /* Over+Undersized */
	hmp->stats.rx_over_errors	+= readl(ioaddr + 0x35C); /* Jabber */
	hmp->stats.rx_crc_errors	+= readl(ioaddr + 0x360);
	hmp->stats.rx_frame_errors	+= readl(ioaddr + 0x364); /* Symbol Errs */
	hmp->stats.rx_missed_errors	+= readl(ioaddr + 0x36C); /* Dropped */

	return &hmp->stats;
}

static void set_rx_mode(struct device *dev)
{
	long ioaddr = dev->base_addr;

	if (dev->flags & IFF_PROMISC) {			/* Set promiscuous. */
		/* Unconditionally log net taps. */
		printk(KERN_NOTICE "%s: Promiscuous mode enabled.\n", dev->name);
		writew(0x000F, ioaddr + AddrMode);
	} else if ((dev->mc_count > 63)  ||  (dev->flags & IFF_ALLMULTI)) {
		/* Too many to match, or accept all multicasts. */
		writew(0x000B, ioaddr + AddrMode);
	} else if (dev->mc_count > 0) { /* Must use the CAM filter. */
		struct dev_mc_list *mclist;
		int i;
		for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
			 i++, mclist = mclist->next) {
			writel(*(u32*)(mclist->dmi_addr), ioaddr + 0x100 + i*8);
			writel(0x20000 | (*(u16*)&mclist->dmi_addr[4]),
				   ioaddr + 0x104 + i*8);
		}
		/* Clear remaining entries. */
		for (; i < 64; i++)
			writel(0, ioaddr + 0x104 + i*8);
		writew(0x0003, ioaddr + AddrMode);
	} else {					/* Normal, unicast/broadcast-only mode. */
		writew(0x0001, ioaddr + AddrMode);
	}
}

#ifdef HAVE_PRIVATE_IOCTL
static int mii_ioctl(struct device *dev, struct ifreq *rq, int cmd)
{
	long ioaddr = dev->base_addr;
	u16 *data = (u16 *)&rq->ifr_data;

	switch(cmd) {
	case SIOCDEVPRIVATE:		/* Get the address of the PHY in use. */
		data[0] = ((struct hamachi_private *)dev->priv)->phys[0] & 0x1f;
		/* Fall Through */
	case SIOCDEVPRIVATE+1:		/* Read the specified MII register. */
		data[3] = mdio_read(ioaddr, data[0] & 0x1f, data[1] & 0x1f);
		return 0;
	case SIOCDEVPRIVATE+2:		/* Write the specified MII register */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		mdio_write(ioaddr, data[0] & 0x1f, data[1] & 0x1f, data[2]);
		return 0;
	case SIOCDEVPRIVATE+3: {
		/* Set rx,tx intr params, from Eric Kasten. */
		u32 *d = (u32 *)&rq->ifr_data;
		writel(d[0], dev->base_addr + TxIntrCtrl);
		writel(d[1], dev->base_addr + RxIntrCtrl);
		printk(KERN_NOTICE "%s: Set interrupt mitigate paramters tx %08x, "
			   "rx %08x.\n", dev->name,
			   (int) readl(dev->base_addr + TxIntrCtrl),
			   (int) readl(dev->base_addr + RxIntrCtrl));
		return 0;
	}
	default:
		return -EOPNOTSUPP;
	}
}
#endif  /* HAVE_PRIVATE_IOCTL */
#ifdef HAVE_CHANGE_MTU
static int change_mtu(struct device *dev, int new_mtu)
{
	if ((new_mtu < 68) || (new_mtu > 1536))
		return -EINVAL;
	if (dev->start)
		return -EBUSY;
	if (debug > 1) printk("%s: Changing MTU to %d.\n", dev->name, new_mtu);
	dev->mtu = new_mtu;
	return 0;
}
#endif


#ifdef MODULE

int init_module(void)
{
	return hamachi_probe(0);
}

void cleanup_module(void)
{
	struct device *next_dev;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	while (root_hamachi_dev) {
		struct hamachi_private *hmp = (void *)(root_hamachi_dev->priv);
		unregister_netdev(root_hamachi_dev);
		iounmap((char *)root_hamachi_dev->base_addr);
		next_dev = hmp->next_module;
		if (hmp->priv_addr)
			kfree(hmp->priv_addr);
		kfree(root_hamachi_dev);
		root_hamachi_dev = next_dev;
	}
}

#endif  /* MODULE */

/*
 * Local variables:
 *  compile-command: "gcc -DMODULE -D__KERNEL__ -Wall -Wstrict-prototypes -O6 -c hamachi.c `[ -f /usr/include/linux/modversions.h ] && echo -DMODVERSIONS`"
 *  SMP-compile-command: "gcc -D__SMP__ -DMODULE -D__KERNEL__ -I/usr/src/linux/net/inet -Wall -Wstrict-prototypes -O6 -c hamachi.c `[ -f /usr/include/linux/modversions.h ] && echo -DMODVERSIONS`"
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 4
 * End:
 */
