/*
 * linux/fs/commit.c
 * 
 * Written by Stephen C. Tweedie <sct@redhat.com>, 1998
 *
 * Copyright 1998 Red Hat corp --- All Rights Reserved
 *
 * This file is part of the Linux kernel and is made available under
 * the terms of the GNU General Public License, version 2, or at your
 * option, any later version, incorporated herein by reference.
 *
 * Journal commit routines for the generic filesystem journaling code;
 * part of the ext2fs journaling system.
 */

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/jfs.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/locks.h>
#include <linux/buffer.h>

/*
 * journal_commit_transaction
 *
 * The primary function for committing a transaction to the log.  This
 * function is called by the journal thread to begin a complete commit.
 */

void journal_commit_transaction(journal_t *journal)
{
	transaction_t *commit_transaction;
	struct buffer_head *bh, *new_bh, *descriptor;
	struct buffer_head *wbuf[journal_flush_nr_buffers];
	int bufs;
	int flags;
	int blocknr;
	char *tagp = NULL;
	journal_header_t *header;
	journal_block_tag_t *tag = NULL, *last_tag;
	int space_left = 0;
	int first_tag = 0;
	int tag_flag;
	
	int i;
	
	/*
	 * First job: lock down the current transaction and wait for 
	 * all outstanding updates to complete. 
	 */
	
	lock_journal(journal);

	J_ASSERT (journal->j_running_transaction != NULL);
	J_ASSERT (journal->j_committing_transaction == NULL);
	
	commit_transaction = journal->j_running_transaction;
	J_ASSERT (commit_transaction->t_state == T_RUNNING);

	jfs_debug (1, "Starting commit of transaction %ld\n",
		   commit_transaction->t_tid);
	
	commit_transaction->t_state = T_LOCKED;
	while (commit_transaction->t_updates != 0) {
		unlock_journal(journal);
		sleep_on(&commit_transaction->t_wait);
		lock_journal(journal);
	}

	J_ASSERT (commit_transaction->t_outstanding_credits <= journal->j_max_transaction_buffers);
	
	/*
	 * First thing we are allowed to do is to discard any remaining
	 * BJ_Reserved buffers.  Note, it is _not_ permissible to assume
	 * that there are no such buffers: if a large filesystem
	 * operation like a truncate needs to split itself over multiple
	 * transactions, then it may try to do a journal_restart() while
	 * there are still BJ_Reserved buffers outstanding.  These must
	 * be released cleanly from the current transaction.
	 *
	 * In this case, the filesystem must still reserve write access
	 * again before modifying the buffer in the new transaction, but
	 * we do not require it to remember exactly which old buffers it
	 * has reserved.  This is consistent with the existing behaviour
	 * that multiple journal_get_write_access() calls to the same
	 * buffer are perfectly permissable.
	 */

	while (commit_transaction->t_reserved_list) {
		bh = commit_transaction->t_reserved_list;
		journal_refile_buffer(bh);
	}
	

	/* 
	 * Start committing the old transaction: time to get a new
	 * running transaction for incoming filesystem updates */

	jfs_debug (3, "commit phase 1\n");

	commit_transaction->t_state = T_FLUSH;
	wake_up(&journal->j_wait_transaction_locked);
	
	journal->j_committing_transaction = commit_transaction;
	journal->j_running_transaction = NULL;
	
	commit_transaction->t_log_start = journal->j_head;

	/* 
	 * Now start flushing things to disk, in the order they appear
	 * on the transaction lists.  Data blocks go first.  
	 */
	
 wait_for_data:
	bufs = 0;
	
	/* Cleanup any flushed data buffers from the data list. */
	journal_clean_data_list(commit_transaction);

	bh = commit_transaction->t_datalist;
	
	if (bh) do {
		if (buffer_dirty(bh) && !buffer_locked(bh)) {
			set_bit(BH_JWrite, &bh->b_state);
			wbuf[bufs++] = bh;
		}
		bh = bh->b_tnext;
	} while (bufs < journal_flush_nr_buffers &&
		 bh != commit_transaction->t_datalist);

	if (bufs) {
		unlock_journal(journal);
		ll_rw_block(WRITE, bufs, wbuf);
		run_task_queue(&tq_disk);
		lock_journal(journal);
		goto wait_for_data;
	}

	/*
	 * If we got through the write loop without submitting any new
	 * IO, then wait for all previously submitted IO on the data
	 * list to complete. 
	 */

	bh = commit_transaction->t_datalist;
	
	if (bh) do {
		clear_bit(BH_JWrite, &bh->b_state);
		if (buffer_locked(bh)) {
			unlock_journal(journal);
			wait_on_buffer(bh);
			lock_journal(journal);
			goto wait_for_data;
		}
		
		bh = bh->b_tnext;
	} while (bh != commit_transaction->t_datalist);

	/*
	 * If we found any dirty or locked buffers, then we should have
	 * looped back up to the wait_for_data label.  If there weren't
	 * any then journal_clean_data_list should have wiped the list
	 * clean by now, so check that it is in fact empty. 
	 */
	
	J_ASSERT (commit_transaction->t_datalist == NULL);

	jfs_debug (3, "Commit phase 2\n");

	/*
	 * Way to go: we have now written out all of the data for a
	 * transaction!  Now comes the tricky part: we need to write out
	 * metadata now.  Loop over the transaction's entire buffer list:
	 */

	commit_transaction->t_state = T_COMMIT;
	descriptor = 0;
	bufs = 0;
	
	while (commit_transaction->t_buffers) {

		/* Find the next buffer to be journaled... */

		bh = commit_transaction->t_buffers;

		/* And make sure we have a descriptor block in which to 
		   record the metadata buffer. */

		if (!descriptor) {
			J_ASSERT (bufs == 0);
			
			jfs_debug(4, "get descriptor\n");

			descriptor = journal_get_descriptor_buffer(journal);
			jfs_debug(4, "JFS: got buffer %ld (%p)\n", 
				  descriptor->b_blocknr, descriptor->b_data);
			header = (journal_header_t *) &descriptor->b_data[0];
			header->h_magic     = htonl(JFS_MAGIC_NUMBER);
			header->h_blocktype = htonl(JFS_DESCRIPTOR_BLOCK);
			header->h_sequence  = htonl(commit_transaction->t_tid);

			tagp = &descriptor->b_data[sizeof(journal_header_t)];
			space_left = descriptor->b_size - sizeof(journal_header_t);
			first_tag = 1;
			set_bit(BH_JWrite, &descriptor->b_state);
			wbuf[bufs++] = descriptor;
			
			/* Record it so that we can wait for IO
                           completion later */
			journal_file_buffer(descriptor, 
					    commit_transaction,
					    BJ_LogCtl);
		}

		/* Where is the buffer to be written? */
		
		blocknr = journal_next_log_block(journal);
		
		/* Make a temporary IO buffer with which to write it out
                   (this will requeue both the metadata buffer and the
                   temporary IO buffer). */

		set_bit(BH_JWrite, &bh->b_state);
		flags = journal_write_metadata_buffer(commit_transaction,
						      bh, 
						      &new_bh,
						      blocknr);
		set_bit(BH_JWrite, &new_bh->b_state);
		wbuf[bufs++] = new_bh;
		
		/* Record the new block's tag in the current descriptor
                   buffer */

		tag_flag = 0;
		if (flags & 1)
			tag_flag |= JFS_FLAG_ESCAPE;
		if (!first_tag)
			tag_flag |= JFS_FLAG_SAME_UUID;

		tag = (journal_block_tag_t *) tagp;
		tag->t_blocknr = htonl(bh->b_blocknr);
		tag->t_flags = htonl(tag_flag);
		tagp += sizeof(journal_block_tag_t);
		space_left -= sizeof(journal_block_tag_t);
		
		if (first_tag) {
			memcpy (tagp, journal->j_uuid, 16);
			tagp += 16;
			space_left -= 16;
			first_tag = 0;
		}
		
		/* If there's no more to do, or if the descriptor is full,
		   let the IO rip! */

		if (bufs == journal_flush_nr_buffers ||
		    commit_transaction->t_buffers == NULL ||		    
		    space_left < sizeof(journal_block_tag_t) + 16) {

			jfs_debug(4, "JFS: Submit %d IOs\n", bufs);
			
			/* Write an end-of-descriptor marker before
                           submitting the IOs.  "tag" still points to
                           the last tag we set up. */
			
			tag->t_flags |= htonl(JFS_FLAG_LAST_TAG);

			unlock_journal(journal);
			ll_rw_block (WRITE, bufs, wbuf);
			lock_journal(journal);
			
			/* Force a new descriptor to be generated next
                           time round the loop. */
			descriptor = NULL;
			bufs = 0;
		}
	}

	/* Lo and behold: we have just managed to send a transaction to
           the log.  Before we can commit it, wait for the IO so far to
           complete.  Control buffers being written are on the
           transaction's t_log_list queue, and metadata buffers are on
           the t_iobuf_list queue. 
	
	   Wait for the transactions in reverse order.  That way we are
	   less likely to be woken up until all IOs have completed, and
	   so we incur less scheduling load.
	*/

	jfs_debug(3, "JFS: commit phase 3\n");

 wait_for_iobuf:
	while (commit_transaction->t_iobuf_list != NULL) {
		struct buffer_head *bh = commit_transaction->t_iobuf_list->b_tprev;
		if (buffer_locked(bh)) {
			unlock_journal(journal); 
			wait_on_buffer(bh);
			lock_journal(journal); 
			goto wait_for_iobuf;
		}
		
		clear_bit(BH_JWrite, &bh->b_state);

		journal_unfile_buffer(bh);
		put_unused_buffer_head(bh);
		
		/* We also have to unlock and free the corresponding
                   shadowed buffer */

		bh = commit_transaction->t_shadow_list->b_tprev;
		clear_bit(BH_JWrite, &bh->b_state);
		J_ASSERT(buffer_dirty(bh));
		
		/* The metadata is now released for reuse, but we need
                   to remember it against this transaction so that when
                   we finally commit, we can do any checkpointing
                   required. */

		journal_file_buffer(bh, commit_transaction, BJ_Forget);

		/* Wake up any transactions which were waiting for this
		   IO to complete */
		wake_up(&bh->b_wait);
		
		/* Found an unlocked buffer: remove it */
		
		if (bh->b_frozen_data) {
			kfree(bh->b_frozen_data);
			bh->b_frozen_data = NULL;
		}
	}

	J_ASSERT (commit_transaction->t_shadow_list == NULL);
	
	jfs_debug(3, "JFS: commit phase 4\n");

 wait_for_ctlbuf:
	while (commit_transaction->t_log_list != NULL) {
		struct buffer_head *bh = commit_transaction->t_log_list->b_tprev;
		
		if (buffer_locked(bh)) {
			unlock_journal(journal); 
			wait_on_buffer(bh);
			lock_journal(journal); 
			goto wait_for_ctlbuf;
		}
		
		clear_bit(BH_JWrite, &bh->b_state);	
		journal_unfile_buffer(bh);
		bh->b_transaction = NULL;
		brelse(bh);
	}
	
	jfs_debug(3, "JFS: commit phase 5\n");

	/* Done it all: now write the commit record */

	descriptor = journal_get_descriptor_buffer(journal);

	for (i = 0; i < descriptor->b_size; i += 512) {
		journal_header_t *tmp = (journal_header_t*)descriptor->b_data;
		tmp->h_magic = htonl(JFS_MAGIC_NUMBER);
		tmp->h_blocktype = htonl(JFS_COMMIT_BLOCK);
		tmp->h_sequence = htonl(commit_transaction->t_tid);
	}

	unlock_journal(journal); 
	ll_rw_block(WRITE, 1, &descriptor);
	wait_on_buffer(descriptor);
	brelse(descriptor);
	lock_journal(journal); 
	
	/* End of a transaction!  Finally, we can do checkpoint
           processing: any buffers committed as a result of this
           transaction can be removed from any checkpoint list it was on
           before. */

	jfs_debug(3, "JFS: commit phase 6\n");

	J_ASSERT(commit_transaction->t_datalist == NULL);
	J_ASSERT(commit_transaction->t_buffers == NULL);
	J_ASSERT(commit_transaction->t_checkpoint_list == NULL);
	J_ASSERT(commit_transaction->t_iobuf_list == NULL);
	J_ASSERT(commit_transaction->t_shadow_list == NULL);
	J_ASSERT(commit_transaction->t_log_list == NULL);

	while (commit_transaction->t_forget) {
		transaction_t *cp_transaction;
		
		bh = commit_transaction->t_forget;
		J_ASSERT (bh->b_transaction == commit_transaction || 
			  bh->b_transaction == journal->j_running_transaction);

		cp_transaction = bh->b_cp_transaction;
		if (cp_transaction) {
			J_ASSERT (commit_transaction != cp_transaction);
			journal_remove_checkpoint(bh);
		}
		journal_insert_checkpoint(bh, commit_transaction);
       
		journal_refile_buffer(bh);
	}
	
	/* Done with this transaction! */

	jfs_debug(3, "JFS: commit phase 7\n");

	J_ASSERT (commit_transaction->t_state == T_COMMIT);
	commit_transaction->t_state = T_FINISHED;

	J_ASSERT (commit_transaction == journal->j_committing_transaction);
	journal->j_commit_sequence = commit_transaction->t_tid;
	journal->j_committing_transaction = NULL;

	if (commit_transaction->t_checkpoint_list == NULL) {
		journal_drop_transaction(journal, commit_transaction);
	} else {
		if (journal->j_checkpoint_transactions == NULL) {
			journal->j_checkpoint_transactions = commit_transaction;
			commit_transaction->t_cpnext = commit_transaction;
			commit_transaction->t_cpprev = commit_transaction;
		} else {
			commit_transaction->t_cpnext = journal->j_checkpoint_transactions;
			commit_transaction->t_cpprev = commit_transaction->t_cpnext->t_cpprev;
			commit_transaction->t_cpnext->t_cpprev = commit_transaction;
			commit_transaction->t_cpprev->t_cpnext = commit_transaction;
		}
	}
	
	jfs_debug(1, "Commit complete\n");

	unlock_journal(journal); 
	wake_up(&journal->j_wait_done_commit);
}
