/*
 * linux/fs/nfs/read.c
 *
 * Block I/O for NFS
 *
 * Partial copy of Linus' read cache modifications to fs/nfs/file.c
 * modified for async RPC by okir@monad.swb.de
 *
 * We do an ugly hack here in order to return proper error codes to the
 * user program when a read request failed: since generic_file_read
 * only checks the return value of inode->i_op->readpage() which is always 0
 * for async RPC, we set the error bit of the page to 1 when an error occurs,
 * and make nfs_readpage transmit requests synchronously when encountering this.
 * This is only a small problem, though, since we now retry all operations
 * within the RPC code when root squashing is suspected.
 */

#define NFS_NEED_NFS2_XDR_TYPES
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>
#include <linux/file.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs.h>
#include <linux/nfs2.h>
#include <linux/nfs3.h>
#include <linux/nfs_fs.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/pgtable.h>

#define NFSDBG_FACILITY		NFSDBG_PAGECACHE

struct nfs_rreq {
	struct inode		*ra_inode;	/* inode to be read */
	unsigned int		ra_npages;	/* number of pages to read */
	struct page		*ra_page[NFS_READ_MAXIOV]; /* pages to read */
	struct nfs_fh		ra_handle;	/* filehandle */
	struct nfs_readargs	ra_args;	/* XDR argument struct */
	struct nfs_readres	ra_res;		/* ... and result struct */
	struct nfs_fattr	ra_fattr;	/* fattr storage */
};

/* Hack for future NFS swap support */
#ifndef IS_SWAPFILE
# define IS_SWAPFILE(inode)	(0)
#endif

static __inline__ struct nfs_rreq *nfs_rreq_alloc(void)
{
	return kmalloc(sizeof(struct nfs_rreq), GFP_KERNEL);
}

static __inline__ void nfs_rreq_free(struct nfs_rreq *p)
{
	kfree(p);
}

/*
 * Set up the NFS read request struct
 */
static inline void
nfs_readreq_setup(struct nfs_rreq *req, struct inode *inode,
		  struct nfs_fh *fh,
		  struct page **pages, unsigned int nr)
{
	int i;

	req->ra_inode	    = inode;
	req->ra_args.fh	    = &req->ra_handle;
	req->ra_handle	    = *fh;
	req->ra_args.offset = page_offset(pages[0]);
	req->ra_args.count  = nr*PAGE_CACHE_SIZE;
	req->ra_args.nriov  = nr;
	req->ra_npages	    = nr;
	for (i = 0; i < nr; i++) {
		req->ra_args.iov[i].iov_base = (void *)page_address(pages[i]);
		req->ra_args.iov[i].iov_len = PAGE_CACHE_SIZE;
		req->ra_page[i] = pages[i];
	}
	req->ra_fattr.valid = 0;
	req->ra_res.fattr   = &req->ra_fattr;
	req->ra_res.count   = req->ra_args.count;
	req->ra_res.eof     = 0;
}


/*
 * Read a page synchronously.
 */
static int
nfs_readpage_sync(struct file *file, struct page *page)
{
	struct dentry	*dentry = file->f_dentry;
	struct inode	*inode = dentry->d_inode;
	struct rpc_cred	*cred = nfs_file_cred(file);
	struct nfs_fattr fattr;
	unsigned long	offset = page_offset(page);
	char		*buffer = (char *) page_address(page);
	int		rsize = NFS_SERVER(inode)->rsize;
	int		result, refresh = 0;
	int		count = PAGE_CACHE_SIZE, chunk, eof = 0;
	int		flags = 0;

	if (IS_SWAPFILE(inode))
		flags |= NFS_RPC_SWAPFLAGS;

	dprintk("NFS: nfs_readpage_sync(%p)\n", page);
	clear_bit(PG_error, &page->flags);

	do {
		if ((chunk = rsize) > count)
			chunk = count;

		dprintk("NFS: nfs_proc_read(%s, (%s/%s), %ld, %d, %p)\n",
			NFS_SERVER(inode)->hostname,
			dentry->d_parent->d_name.name, dentry->d_name.name,
			offset, chunk, buffer);

		result = NFS_CALL(read, inode, (dentry, &fattr, cred, flags,
						offset, chunk, buffer, &eof));
		nfs_refresh_inode(inode, &fattr);

		/*
		 * Even if we had a partial success we can't mark the page
		 * cache valid.
		 */
		if (result < 0) {
			if (result == -EISDIR)
				result = -EINVAL;
			goto io_error;
		}
		refresh = 1;
		count  -= result;
		offset += result;
		buffer += result;
		if (eof)
			break;
	} while (count);

	memset(buffer, 0, count);
	flush_dcache_page(page_address(page));
	set_bit(PG_uptodate, &page->flags);
	result = 0;

io_error:
	/* Note: we don't refresh if the call returned error */
	if (result < 0)
		set_bit(PG_error, &page->flags);
	return result;
}

struct page *
nfs_grab_cache_page(struct inode *inode, unsigned long offset)
{
	struct page	*page,
			**hash;
	unsigned long	new_addr = 0;

	offset &= PAGE_CACHE_MASK;
	hash = page_hash(inode, offset);
	do {
		page = __find_page(inode, offset, *hash);
		if (page) {
			if (new_addr)
				page_cache_free(new_addr);
			break;
		}
		if (!new_addr) {
			new_addr = page_cache_alloc();
			if (new_addr)
				continue;
			return NULL;
		}
		page = page_cache_entry(new_addr);
		add_to_page_cache(page, inode, offset, hash);
		new_addr = 0;
	} while (!page);
	while(!nfs_lock_page(page))
		wait_on_page(page);
	return page;
}

/* Page Readahead. We try to allocate and/or lock up to 'nr' empty
 * pages that are to be read in by the same RPC call.
 *
 * FIXME: Allocating pages and handing them down to the nfs_read
 * routine really should be done by generic_file_read.  */
static int
nfs_grab_readahead_pages(struct inode *inode, struct page **pages,
			 unsigned long offset, int nr)
{
	struct page	*page;
	int		error, count = 0;

	offset &= PAGE_CACHE_MASK;
	for (; nr; offset += PAGE_CACHE_SIZE, nr--) {
		if (offset >= inode->i_size)
			break;
		page = nfs_grab_cache_page(inode, offset);
		if (!page)
			break;
		error = nfs_wb_page(inode, page);
		if (error < 0 || PageUptodate(page)) {
			nfs_unlock_page(page);
			page_cache_release(page);
			break;
		}
		pages[count++] = page;
	}
	return count;
}

/*
 * This is the callback from RPC telling us whether a reply was
 * received or some error occurred (timeout or socket shutdown).
 */
static void
nfs_readpage_result(struct rpc_task *task)
{
	struct nfs_rreq	*req = (struct nfs_rreq *) task->tk_calldata;
	struct inode	*inode = req->ra_inode;
	struct page	**pages = req->ra_page;
	int		i, result = task->tk_status;
	int		count, npages = req->ra_npages;

	dprintk("NFS: %4d received callback for %d pages %lx, result %d\n",
			task->tk_pid, npages, page_address(*pages), result);

	nfs_refresh_inode(inode, &req->ra_fattr);

	i = 0;
	if (result >= 0) {
		count = req->ra_res.count;
		/* NFSv3 may return less data than requested,
		 * and have the client issue another request for
		 * the remaining data but only if the client sets
		 * a too large rsize.
		 */
		if (!req->ra_res.eof && result < npages*PAGE_CACHE_SIZE) {
			printk(KERN_WARNING
				"NFS: server sends short reads. "
				"Expected %lu, got %u. "
				"Try reducing the value of rsize.\n",
				PAGE_CACHE_SIZE, result);
		}

		/*
		 * Refresh before we set PG_uptodate in case of invalidation
		 */
		for (; i < npages && count > 0; i++, count -= PAGE_CACHE_SIZE) {
			struct page *page = pages[i];
			if (count < PAGE_CACHE_SIZE)
				memset((char *) page_address(page) + count, 0,
				       PAGE_CACHE_SIZE - count);
			set_bit(PG_uptodate, &page->flags);
		}
	}
	for (; i < npages; i++)
		set_bit(PG_error, &pages[i]->flags);

	/* N.B. Use nfs_unlock_page here? */
	for (i = 0; i < npages; i++) {
		nfs_unlock_page(pages[i]);
		page_cache_release(pages[i]);
	}

	nfs_rreq_free(req);
}

static inline int
nfs_readpage_async(struct file *file, struct page *page)
{
	struct dentry	*dentry = file->f_dentry;
	struct inode	*inode = dentry->d_inode;
	struct rpc_message msg;
	struct nfs_rreq	*req;
	struct page	*pages[NFS_READ_MAXIOV];
	int		result = -ENOMEM,
			i, flags, rpages;

	dprintk("NFS: nfs_readpage_async(%p)\n", page);

	/* N.B. Do we need to test? Never called for swapfile inode */
	flags = RPC_TASK_ASYNC;

	if (IS_SWAPFILE(inode))
		flags |= NFS_RPC_SWAPFLAGS;

	req = nfs_rreq_alloc();
	if (!req)
		goto out_defer;

	pages[0] = page;
	rpages = 1;
	if (NFS_SERVER(inode)->rpages > 1) {
		int npages = NFS_SERVER(inode)->rpages - 1;
		if (npages > NFS_READ_MAXIOV - 1)
			npages = NFS_READ_MAXIOV - 1;

		rpages += nfs_grab_readahead_pages(inode, &pages[1], page->offset+PAGE_CACHE_SIZE, npages);
	}

	/* Initialize request */
	/* N.B. Will the dentry remain valid for life of request? */
	nfs_readreq_setup(req, inode, NFS_FH(dentry), pages, rpages);

	/* Start the async call */
	dprintk("NFS: executing async READ request.\n");

	msg.rpc_proc = (NFS_PROTO(inode)->version == 3) ? NFS3PROC_READ : NFSPROC_READ;
	msg.rpc_argp = &req->ra_args;
	msg.rpc_resp = &req->ra_res;
	msg.rpc_cred = nfs_file_cred(file);

	result = rpc_call_async(NFS_CLIENT(inode), &msg, flags,
				nfs_readpage_result, req);
	if (result < 0)
		goto out_free;
	result = 0;
 out:
	return result;

 out_defer:
	dprintk("NFS: deferring async READ request.\n");
	goto out;
 out_free:
	dprintk("NFS: failed to enqueue async READ request.\n");
	for(i = 1; i < rpages; i++) {
		set_bit(PG_error, &pages[i]->flags);
		nfs_unlock_page(pages[i]);
		page_cache_release(pages[i]);
	}
	nfs_rreq_free(req);
	goto out;
}

/*
 * Read a page over NFS.
 * We read the page synchronously in the following cases:
 *  -	The file is a swap file. Swap-ins are always sync operations,
 *	so there's no need bothering to make async reads 100% fail-safe.
 *  -	The NFS rsize is smaller than PAGE_CACHE_SIZE. We could kludge our way
 *	around this by creating several consecutive read requests, but
 *	that's hardly worth it.
 *  -	The error flag is set for this page. This happens only when a
 *	previous async read operation failed.
 *  -	The server is congested.
 */
int
nfs_readpage(struct file *file, struct page *page)
{
	struct dentry *dentry = file->f_dentry;
	struct inode *inode = dentry->d_inode;
	int		error = 0,
			rsize = NFS_SERVER(inode)->rsize,
			rpages = NFS_SERVER(inode)->rpages;

	atomic_inc(&page->count);
	while (!nfs_lock_page(page))
		wait_on_page(page);

	dprintk("NFS: nfs_readpage (%p %d@%ld)\n",
		page, rsize, page->offset);

	/*
	 * Try to flush any pending writes to the file
	 */
	error = nfs_sync_file(inode, 0, page_index(page), rpages, FLUSH_WAIT|FLUSH_STABLE);
	if (error < 0)
		goto out_unlock;

	error = -1;
	if (!IS_SWAPFILE(inode) && !PageError(page) && rsize >= PAGE_CACHE_SIZE)
		error = nfs_readpage_async(file, page);

	if (error >= 0)
		goto out;

	error = nfs_readpage_sync(file, page);
	if (error < 0 && IS_SWAPFILE(inode))
		printk(KERN_ERR "Aiee.. nfs swap-in of page failed!\n");

 out_unlock:
	nfs_unlock_page(page);
	page_cache_release(page);
 out:
	return error;
}
